# Copyright 2014, Red Hat, Inc
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors:
#    Josef Skladanka <jskladan@edhat.com>
#    Lukas Brabec <lbrabec@redhat.com>

import itertools

from flask import request, jsonify

from oraculum import app, CACHE
from oraculum.action_providers import ACTION_PROVIDERS
from oraculum.utils import blockerbugs, fedocal, meetbot, schedule


def api_v1_landing_page():
    releases = CACHE.get("fedora_releases")
    return {
        'blockerbugs': blockerbugs.get_blockerbugs(),
        'meetings': fedocal.get_qa_meetings(),
        'last_qa_meeting': meetbot.get_last_qa_meeting(),
        'schedule': schedule.get_schedule(),
        'stable': releases["stable"],
        'devel': releases["branched"] or releases["rawhide"],
    }


@app.route('/api/v1/landing_page')
def route_api_v1_landing_page():
    return jsonify(CACHE.get('landing_page'))


def get_actions(provider=None, tags=None):
    actions = []
    for p_name, p_module in ACTION_PROVIDERS.items():
        actions.extend(CACHE.get(p_name))

    if provider:
        actions = [a for a in actions if a['provider'] == provider]
    if tags:
        tags = set(tags.split(','))
        actions = [a for a in actions if tags.intersection(a['tags'])]
    return actions


@app.route('/api/v1/actions')
def route_api_v1_actions():
    return jsonify(api_v1_actions())


def api_v1_actions():
    provider = request.args.get('provider')
    tags = request.args.get('tags')

    return {'actions': get_actions(provider, tags)}


@app.route('/api/v1/actions/providers')
def api_v1_actions_providers():
    resp = {'providers': [p.INFO for p in ACTION_PROVIDERS.values()]}
    return jsonify(resp)


@app.route('/api/v1/actions/tags')
def api_v1_actions_tags():
    provider = request.args.get('provider')
    actions = get_actions(provider)

    resp = {'tags': []}
    if actions:
        resp['tags'] = list(set(itertools.chain(*[a['tags'] for a in actions])))

    return jsonify(resp)


@app.route('/api/v1/actions/all')
def api_v1_actions_all():

    resp = {
        'providers': [p.INFO for p in ACTION_PROVIDERS.values()],
        'actions': {p: get_actions(p) for p in ACTION_PROVIDERS.keys()}
    }

    return jsonify(resp)
