# Copyright 2014, Red Hat, Inc
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors:
#    Frantisek Zatloukal <fzatlouk@redhat.com>
#    Josef Skladanka <jskladan@redhat.com>

from flask import jsonify

from oraculum import app, CACHE
from oraculum.utils import cache_utils, orphans, health_check, bodhi, koschei, pagure, dashboard_helpers

def handle_orphan_user():
    orphans_data = list(CACHE.get('orphans_json')['affected_packages'].keys())
    pkg_dict = {package: [] for package in orphans_data}
    empty = {'status': 200, 'data': pkg_dict}
    static_info = {
        'packages': orphans_data,
        'group_packages': [],
        'primary_packages': orphans_data,
        'orphans': {},
        'fails_to_install': pkg_dict,
        'updates': pkg_dict,
        'overrides': pkg_dict,
        'koschei': pkg_dict
        }

    static_info['orphans'] = orphans.get_orphans(static_info['packages'])

    return {
        'static_info': {"data": static_info, 'status': 200, 'last_synced': CACHE.get_refreshed_time('orphans_json').isoformat()},
        'prs': empty,
        'bzs': empty,
        'package_versions': empty
        }

@app.route('/api/v1/packager_dashboard_caching')
def data_caching_condition():
    return jsonify({"visits_required_every_n_days": app.config["ACTIVITY_REQUIRED"]})

@app.route('/api/v1/packager_dashboard/<user>', methods=['GET'])
def route_dashboard_user_data(user):
    if user == 'orphan':
        return jsonify(CACHE.get('packager-dashboard_orphan_page'))

    # Cleanup the username to follow FAS naming restrictions
    user = dashboard_helpers.clean_fas_username(user)

    packages_promise = CACHE.async_get('packager-dashboard_user_data_static', 'high', user)
    last_synced = CACHE.get_refreshed_time('packager-dashboard_user_data_static', user)
    if packages_promise == cache_utils.RefresherNotRegistered:
        static_info = {'status': 404, 'data': None, 'last_synced': None}
    elif packages_promise == cache_utils.AsyncRefreshInProgress:
        static_info = {'status': 204, 'data': None, 'last_synced': None}
    else:
        static_info = {'status': 200, 'data': packages_promise, 'last_synced': last_synced.isoformat()}
        if len(static_info["data"]["packages"]) > 0:
            dashboard_helpers.update_user_access_time(user)

    prs = dashboard_user_data_prs(user)
    bzs = dashboard_user_data_bzs(user)
    package_versions = dashboard_user_data_pkg_versions(user)

    return jsonify({'static_info': static_info, 'prs': prs, 'bzs': bzs, 'package_versions': package_versions})


def dashboard_user_data_static(user):
    packages = pagure.get_packages(user, CACHE.get('packages_owners_json'), CACHE.get('pagure_groups'))
    # Just throw out empty lists and dicts for users without any packages
    if len(packages["combined"]) == 0:
        return {
            'packages': [],
            'group_packages': {},
            'primary_packages': [],
            'orphans': {},
            'fails_to_install': {},
            'updates': {},
            'overrides': {},
            'koschei': {}
        }
    return {
        'packages': packages["combined"],
        'group_packages': packages["group"],
        'primary_packages': packages["primary"],
        'orphans': orphans.get_orphans(packages["combined"]),
        'fails_to_install': health_check.get_health_check_user_data(packages["combined"]),
        'updates': bodhi.get_updates(packages["combined"], CACHE.get('bodhi_updates')),
        'overrides': bodhi.get_user_overrides(packages["combined"], CACHE.get('bodhi_overrides')),
        'koschei': koschei.get_user_koschei_data(packages["combined"]),
    }


def dashboard_user_data_pkg_versions(user):
    data = CACHE.async_get('packager-dashboard_user_data_static', 'low', user)
    if data in [cache_utils.RefresherNotRegistered, cache_utils.AsyncRefreshInProgress]:
        return {"status": 204, "data": None}

    status = 200
    data, load_status = CACHE.async_get_batch('packager-dashboard_package_versions', data["packages"], 'low')
    if load_status in [cache_utils.RefresherNotRegistered, cache_utils.AsyncRefreshInProgress]:
        status = 202
    return {"status": status, "data": data}


def dashboard_user_data_prs(user):
    data = CACHE.async_get('packager-dashboard_user_data_static', 'low', user)
    if data in [cache_utils.RefresherNotRegistered, cache_utils.AsyncRefreshInProgress]:
        return {"status": 204, "data": None}

    status = 200
    data, load_status = CACHE.async_get_batch('packager-dashboard_pull_requests', data["packages"], 'low')
    if load_status in [cache_utils.RefresherNotRegistered, cache_utils.AsyncRefreshInProgress]:
        status = 202
    return {"status": status, "data": data}


def dashboard_user_data_bzs(user):
    data = CACHE.async_get('packager-dashboard_user_data_static', 'low', user)
    if data in [cache_utils.RefresherNotRegistered, cache_utils.AsyncRefreshInProgress]:
        return {"status": 204, "data": None}

    status = 200
    data, load_status = CACHE.async_get_batch('packager-dashboard_bugs', data["packages"], 'low')
    if load_status in [cache_utils.RefresherNotRegistered, cache_utils.AsyncRefreshInProgress]:
        status = 202
    return {"status": status, "data": data}
