#
# cache_utils.py - Database helper functions landing page cache
#
# Copyright 2020, Red Hat, Inc
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors:
#   Josef Skladanka <jskladan@redhat.com>

import datetime
import json

import oraculum
from oraculum import app, celery_app, db
from oraculum.models.db_cache import CachedData
from oraculum.models.watchdog import WatchDogData
from oraculum.utils import celery_utils, watchdog_utils

from sqlalchemy import exc


class RefresherNotRegistered():
    pass


class AsyncRefreshInProgress():
    pass


class CachedObject():
    def __init__(self, time_created, data):
        self.time_created = time_created
        self.data = data


class Cached():
    def __init__(self, max_cache_age):
        self._refreshers = {}
        self._local_cache = {}
        self._max_cache_age = datetime.timedelta(seconds=max_cache_age)

    # getting the values
    @property
    def max_cache_age(self):
        return self._max_cache_age

    # setting the values
    @max_cache_age.setter
    def max_cache_age(self, value):
        self._max_cache_age = datetime.timedelta(seconds=value)

    def register(self, what, refresh_method):
        self._refreshers[what] = refresh_method

    def _construct_what(self, item, *args, **kwargs):
        return '%s [%s, %s]' % (item, args, kwargs)

    def get_refreshed_time(self, item, *args, **kwargs):
        what = self._construct_what(item, *args, **kwargs)

        local = self._local_cache.get(what)
        if self._new_enough(local):
            return local.time_created

        from_db = CachedData.query.filter_by(provider=what).first()
        if self._new_enough(from_db):
            return from_db.time_created

        return None

    def __get(self, synchronous, priority, item, *args, **kwargs):
        if item not in self._refreshers:
            return RefresherNotRegistered

        what = self._construct_what(item, *args, **kwargs)

        if self._new_enough(self._local_cache.get(what)):
            return self._local_cache[what].data

        #app.logger.debug("local cache cache miss on %s" % what)
        from_db = CachedData.query.filter_by(provider=what).first()

        if from_db and self._new_enough(from_db):
            if app.config['ENABLE_LOCAL_CACHE']:
                self._local_cache[what] = CachedObject(from_db.time_created, from_db.data)
            return from_db.data

        #app.logger.debug("db cache miss on %s" % what)
        if synchronous:
            return self._refresh(item, *args, **kwargs)

        celery_utils.plan_celery_refresh(priority, item, *args, **kwargs)
        return AsyncRefreshInProgress

    def async_get(self, what, priority="medium", *args, **kwargs):
        return self.__get(False, priority, what, *args, **kwargs)

    def async_get_batch(self, what_base, items, priority="medium"):
        whats = set()
        what_to_item = {}
        data = {}

        if what_base not in self._refreshers:
            return (None, RefresherNotRegistered)

        for item in items:
            what = self._construct_what(what_base, item)

            if app.config['ENABLE_LOCAL_CACHE']:
                cached = self._local_cache.get(what)
                if self._new_enough(cached):
                    data[item] = cached.data
                    continue

            whats.add(what)
            what_to_item[what] = item

        if not whats:
            return (data, None)

        from_db = CachedData.query.filter(CachedData.provider.in_(list(whats))).all()

        # We need to remove the results from the session so we don't get sqlalchemy.exc.UnboundExecutionError later
        db.session.expunge_all()

        seen = set()
        for result in from_db:
            if not self._new_enough(result):
                continue
            item = what_to_item[result.provider]
            data[item] = result.data
            seen.add(item)
            if app.config['ENABLE_LOCAL_CACHE']:
                self._local_cache[result.provider] = CachedObject(result.time_created, result.data)

        # Schedule refresh for any missing result that we've asked for
        missing = set(items).difference(seen)

        for missed in missing:
            celery_utils.plan_celery_refresh(priority, what_base, missed)
            data[missed] = []

        if missing:
            return (data, AsyncRefreshInProgress)
        return (data, None)

    def get(self, what, *args, **kwargs):
        return self.__get(True, "medium", what, *args, **kwargs)

    def _new_enough(self, cached_object):
        if cached_object is None:
            return False
        if app.config['IGNORE_CACHE_TIMESTAMP'] == "local":
            return True
        # If cached_object is an instance of CachedData, it is from the database
        if app.config['IGNORE_CACHE_TIMESTAMP'] == "db" and isinstance(cached_object, CachedData):
            return True
        if datetime.datetime.utcnow() > cached_object.time_created + self.max_cache_age:
            return False
        return True

    def _refresh(self, item, *args, **kwargs):
        what = self._construct_what(item, *args, **kwargs)
        data = self._refreshers[item](*args, **kwargs)
        if data is None:
            app.logger.debug("Refresher for %s returned non-data %r" % (what, data))
            return None

        row = CachedData.query.filter_by(provider=what).first()
        if row:
            row.time_created = datetime.datetime.utcnow()
            row.raw_text = json.dumps(data)
        else:
            row = CachedData(what, json.dumps(data))
            db.session.add(row)

        try:
            db.session.commit()
        except exc.IntegrityError:
            # User probably refreshed page during loading and we hit IntegrityError
            # Try to rollback queued changes
            db.session.rollback()
            row = CachedData.query.filter_by(provider=what).first()
            data = row.data

        # Store time of sync finish
        watchdog_utils.push_to_watchdog("sync_ended", what)

        if app.config['ENABLE_LOCAL_CACHE']:
            self._local_cache[what] = CachedObject(row.time_created, data)
        return data


if __name__ == "__main__":
    CACHE = Cached(max_cache_age=300)
    import ipdb
    ipdb.set_trace()
    """
    CACHE.register('bodhi', libkarma.get_all_updates)

    print("- Getting bodhi")
    test = CACHE.get('bodhi')
    """
