#
# celery_utils.py - Functions for oraculum/celery integration
#
# Copyright 2020, Red Hat, Inc
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors:
#   Frantisek Zatloukal <fzatlouk@redhat.com>
#   Josef Skladanka <jskladan@redhat.com>

import sys
import datetime
import traceback

import oraculum
from oraculum import app, celery_app
from oraculum.action_providers import ACTION_PROVIDERS
from oraculum.models.dashboard_users import DashboardUserData
from oraculum.utils.watchdog_utils import push_to_watchdog, process_queue

@celery_app.task
def plan_celery_refresh(priority, *args, **kwargs):
    # Priority and redis
    # used to be lower number-lower priority as per https://docs.celeryproject.org/projects/kombu/en/latest/changelog.html#version-4-0
    # changed (reversed) in 4.0.2 as per: https://docs.celeryproject.org/projects/kombu/en/latest/changelog.html#version-4-0-2
    priorities = {'low': 9,
                  'medium': 5,
                  'high': 3}
    priority = priorities.get(priority, priorities['medium'])

    what = oraculum.CACHE._construct_what(*args, **kwargs)

    redis_conn = celery_app.broker_connection().default_channel.client
    counter = redis_conn.incr(what)
    if counter == 1:
        app.logger.debug("PLANNING - celery_refresh on %s" % what)
        push_to_watchdog("sync_planned", what)
        celery_refresh.apply_async(args=args, kwargs=kwargs, priority=priority)
    else:
        app.logger.debug("celery_refresh already planned for %s " % what)

@celery_app.task
def celery_refresh(*args, **kwargs):
    what = oraculum.CACHE._construct_what(*args, **kwargs)
    app.logger.debug("START - celery_refresh on %s" % what)
    # run refresher
    try:
        push_to_watchdog("sync_started", what)
        oraculum.CACHE._refresh(*args, **kwargs)
    except Exception as e:
        app.logger.error("ERROR - celery_refresh on %s\n%s" % (what, e))
        push_to_watchdog("sync_failed", what, traceback.format_exc())
        #raise
    finally:
        # Cleanup counter to allow another scheduling
        redis_conn = celery_app.broker_connection().default_channel.client
        counter = redis_conn.getset(what, 0)

    app.logger.debug("END - celery_refresh on %s" % what)



@celery_app.on_after_configure.connect
def setup_periodic_tasks(sender, **kwargs):
    app.logger.debug('@celery_app.on_after_configure.connect - Registering cache refreshers')

    oraculum.main.register_cache_providers()

    if not (('beat' in sys.argv) or ('worker' in sys.argv and '-B' in sys.argv)):
        app.logger.debug('@celery_app.on_after_configure.connect - Non beat process detected')
        return

    SYNC_INTERVALS = app.config['SYNC_INTERVALS']

    #app.logger.debug("Adding periodic task heart_beat every %s seconds" % SYNC_INTERVALS["heart_beat"])
    #sender.add_periodic_task(SYNC_INTERVALS["heart_beat"], heart_beat.s(), priority = 1)

    # WatchDog
    sender.add_periodic_task(app.config["WATCHDOG_INTERVAL"], process_queue.s(), priority=0)
    app.logger.debug("Scheduling WatchDog every %s seconds" % app.config["WATCHDOG_INTERVAL"])

    # Landing page
    app.logger.debug("Adding periodic task landing_page every %s seconds" % SYNC_INTERVALS["landing_page"])
    sender.add_periodic_task(SYNC_INTERVALS["landing_page"], plan_celery_refresh.s("medium", "landing_page"), priority=0)
    for p_name, p_module in ACTION_PROVIDERS.items():
        app.logger.debug("Adding periodic task %s every %s seconds" % (p_name, SYNC_INTERVALS[p_name]))
        sender.add_periodic_task(SYNC_INTERVALS[p_name], plan_celery_refresh.s("medium", p_name), priority=0)

    # Packager Dashboard - Generic data same for all users
    GENERIC_SYNCERS = ['pagure_groups', 'packages_owners_json', 'orphans_json', 'koschei_data',
                       'bodhi_overrides', 'bodhi_updates', 'ftbfs_trackers', 'fti_trackers',
                       'health_check_data', 'blocker_trackers', 'fedora_releases']

    for syncer in GENERIC_SYNCERS:
        app.logger.debug("Adding periodic task %s every %s seconds" % (syncer, SYNC_INTERVALS[syncer]))
        sender.add_periodic_task(SYNC_INTERVALS[syncer], plan_celery_refresh.s('medium', syncer), priority=0)

    # Let's cache also data for orphan page
    app.logger.debug("Adding periodic task %s every %s seconds" % ("packager-dashboard_orphan_page", SYNC_INTERVALS["orphans_json"]))
    sender.add_periodic_task(SYNC_INTERVALS["orphans_json"], plan_celery_refresh.s('medium', 'packager-dashboard_orphan_page'), priority=0)

    # Packager Dashboard - Sync data for users who meet criteria for cached access to the site
    app.logger.debug("Adding periodic task packager_dashboard_user_data every %s seconds" %
                     SYNC_INTERVALS["packager_dashboard_static_user_data"])
    sender.add_periodic_task(SYNC_INTERVALS["packager_dashboard_static_user_data"], celery_sync_static_user_data.s(), priority=2)
    app.logger.debug("Adding periodic task to sync package bugs every %s seconds" %
                     SYNC_INTERVALS["packager_dashboard_package_bugs"])
    sender.add_periodic_task(SYNC_INTERVALS["packager_dashboard_package_bugs"], celery_sync_package_bugs.s(), priority=2)
    app.logger.debug("Adding periodic task to sync package prs every %s seconds" %
                     SYNC_INTERVALS["packager_dashboard_package_prs"])
    sender.add_periodic_task(SYNC_INTERVALS["packager_dashboard_package_prs"], celery_sync_package_prs.s(), priority=2)
    app.logger.debug("Adding periodic task to sync package versions every %s seconds" %
                     SYNC_INTERVALS["packager_dashboard_package_versions"])
    sender.add_periodic_task(SYNC_INTERVALS["packager_dashboard_package_versions"], celery_sync_package_versions.s(), priority=2)


@celery_app.task
def heart_beat():
    app.logger.debug("Heart Beat....")


def get_users_for_sync():
    """
    Returns list of usernames to be included in sync if conditions specified in settings.py are met
    """
    now = datetime.datetime.utcnow() - datetime.timedelta(days=app.config['ACTIVITY_REQUIRED'])
    row = DashboardUserData.query.filter(DashboardUserData.last_accessed >= now)
    return [user.username for user in row]


@celery_app.task
def celery_sync_static_user_data():
    users = get_users_for_sync()
    for user in users:
        plan_celery_refresh("medium", 'packager-dashboard_user_data_static', user)

@celery_app.task
def celery_sync_package_bugs():
    users = get_users_for_sync()
    for user in users:
        data = oraculum.controllers.packager_dashboard.dashboard_user_data_static(user)
        for package in data["packages"]:
            plan_celery_refresh("low", 'packager-dashboard_bugs', package)

@celery_app.task
def celery_sync_package_prs():
    users = get_users_for_sync()
    for user in users:
        data = oraculum.controllers.packager_dashboard.dashboard_user_data_static(user)
        for package in data["packages"]:
            plan_celery_refresh("low", 'packager-dashboard_pull_requests', package)

@celery_app.task
def celery_sync_package_versions():
    users = get_users_for_sync()
    for user in users:
        data = oraculum.controllers.packager_dashboard.dashboard_user_data_static(user)
        for package in data["packages"]:
            plan_celery_refresh("low", 'packager-dashboard_package_versions', package)
