#
# health_check.py - Health Check parsers and utilities for packager dashboard
#
# Copyright 2020, Red Hat, Inc
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors:
#   Frantisek Zatloukal <fzatlouk@redhat.com>
#   Josef Skladanka <jskladan@redhat.com>

from collections import defaultdict

from oraculum import app, CACHE

from oraculum.utils.dashboard_helpers import get_json, release_from_number


def process_health_check_json(json_data):
    """
    Cleans up data for further health check processing into following format:
    {"package": {"arch": ["reason_a", "reason_b"], "arch_b": ["reason_a, reason_c"]}, "package_b"...}
    """
    data = defaultdict(dict)
    for single_closure in json_data:
        if single_closure["arch"] == "src":
            arch = "src-" + single_closure["repo_arch"]
        else:
            arch = single_closure["repo_arch"]
        data[single_closure["package"]][arch] = {
            "reason": single_closure["broken"],
            "since": single_closure["since"]
        }
    return data


def get_health_check_data():
    """
    Retrieves json files containing data from fedora-health-check and combines them into dictionary with per-release keys
    Returns dict {
        "FXX":         {"package": {"arch": ["reason_a", "reason_b"], "arch_b": ["reason_a, reason_c"]}, "package_b"...}
        "FXX-testing": {"package": {"arch": ["reason_a", "reason_b"], "arch_b": ["reason_a, reason_c"]}, "package_b"...}
        #Rawhide releases don't have -testing counterpart
        "rawhide":     {"package": {"arch": ["reason_a", "reason_b"], "arch_b": ["reason_a, reason_c"]}, "package_b"...}
    }
    Reason format is described below in get_health_check_user_data
    """
    data = {}
    releases = CACHE.get('fedora_releases')["values"]
    for release in releases[:-1]: # Process every release but rawhide
        release = str(release)
        # Remote resource might not have data for some Fedora Release, typically branched right after branching point
        try:
            data[release] = process_health_check_json(get_json(app.config["HEALTH_CHECK_URL"] + release))
            data[release + "-testing"] = process_health_check_json(get_json(app.config["HEALTH_CHECK_URL"] + release + "-testing"))
        except TypeError:
            pass
    try:
        data["rawhide"] = process_health_check_json(get_json(app.config["HEALTH_CHECK_URL"] + "rawhide"))
    except TypeError:
        pass
    return data


def get_repo_for_health_check(release):
    """
    Used to give frontend a hint if a failure comes from stable, testing or rawhide repos
    """
    if "-testing" in release:
        return "testing"
    if "rawhide" in release:
        return "rawhide"
    return "stable"


def get_health_check_user_data(packages):
    """
    Parses data from get_health_check_data and returns dict for package list:
    {
        "package_a":  {"fails_to_install": True/False, "releases": [{"release": XX, "problems": problems}]
        ...
    }
    "problems": {
        "arch_a": {"reason": ["package_a",...], "since": "2020-07-17T07:47:25.219142397Z"}
        "arch_b": {"reason": ["package_a",...], "since": "2040-07-17T07:47:25.219142397Z"}
        ...
    }
    """
    fti_data = CACHE.get('health_check_data')
    data = {}
    for package in packages:
        data[package] = []
        for release in fti_data:
            if package in fti_data[release]:
                data[package].append({"release": release_from_number(release.replace("-testing", "")), # We need to cut -testing,
                                      "repo": get_repo_for_health_check(release),
                                      "problems": fti_data[release][package]})
    return data
