#
# pagure.py - Pagure parsers and utilities for packager dashboard
#
# Copyright 2020, Red Hat, Inc
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors:
#   Frantisek Zatloukal <fzatlouk@redhat.com>
#   Josef Skladanka <jskladan@redhat.com>

import datetime
import re

from oraculum.utils.dashboard_helpers import get_json, release_from_dist

from oraculum import app, CACHE


def get_pr_ci_result(package, pr_id):
    """
    Return CI result for package pull-request
    Returns {"ci-name": "result", "different-ci-name": "result",...} ,
    "result" can be whatever a CI returns, there is no damned standard for it
    "ci-name" can be anything, Zuul and simple-koji-ci will be the most common at the time of writing the code
    """
    resp_pr_ci_result = get_json("https://src.fedoraproject.org/api/0/rpms/%s/pull-request/%s/flag" % (package, pr_id))
    data = {}
    if not resp_pr_ci_result:
        return {}
    if "flags" not in resp_pr_ci_result:
        return {}
    for result in resp_pr_ci_result["flags"]:
        # We care only about the newest results
        if result["username"] not in data.keys():
            data[result["username"]] = result["status"]
    return data


def get_package_prs(package):
    """
    Returns all open Pull Requests for a single package
    """
    data = []
    resp_package_prs = get_json("https://src.fedoraproject.org/api/0/rpms/%s/pull-requests" % package)
    if not resp_package_prs:
        return None
    # Filter out packages that, for some reason, do not exist in dist-git
    if "error_code" in resp_package_prs and resp_package_prs["error_code"] == "ENOPROJECT":
        return []
    if not "requests" in resp_package_prs:
        app.logger.error("Skipping PRs from package %s because Pagure returned invalid data" % package)
        return []  # Return early if pagure sent invalid data
    for request in resp_package_prs["requests"]:
        if request["status"] == "Open":
            data.append({
                "title": request["title"],
                "author": request["user"]['name'],
                "comments": len(request["comments"]),
                "date_created": str(datetime.datetime.fromtimestamp(int(request["date_created"]))),
                "last_updated": str(datetime.datetime.fromtimestamp(int(request["last_updated"]))),
                "ci_status": get_pr_ci_result(package, request["id"]),
                "release": "Fedora Rawhide" if request["branch"] == "master" else release_from_dist(request["branch"]),
                "url": "https://src.fedoraproject.org/rpms/%s/pull-request/%s" % (package, request["id"])
            })
    return data

def get_pagure_groups():
    """
    Returns dictionary mapping all packager groups in pagure, it's members and packages
    "group_name" : {
        "users": [user_a, ...],
        "packages": [package_a, ...]
    }
    """
    groups_users = {}
    resp = get_json("https://src.fedoraproject.org/api/0/groups?per_page=100")  # TODO: Handle pagination properly
    if not resp:
        return None
    allowed_groups = [re.compile(a) for a in app.config["ALLOWED_PACKAGER_GROUPS"]]
    for group in resp["groups"]:
        if not any(regex.match(group) for regex in allowed_groups):
            continue
        app.logger.debug("Checking out Pagure group %s" % group)
        group_resp = get_json("https://src.fedoraproject.org/api/0/group/%s?projects=1&acl=commit" % group)
        if not group_resp:
            continue
        try:
            groups_users[group] = {
                "users": group_resp["members"],
                "packages": [project["name"] for project in group_resp["projects"]]
            }
        except TypeError:
            app.logger.error("Skipped Pagure group %s because of an error" % group)
            continue
    return groups_users


def get_user_group_packages(user, groups_map):
    """
    Returns list of packages user owns through a group
    """
    group_packages_dict = {}
    group_packages_plain = set()
    for group in groups_map:
        if group == "packager":  # It seems every packager is in group "packager" and inherits zero packages through it
            continue
        if user in groups_map[group]["users"]:
            group_packages_dict[group] = groups_map[group]["packages"]
            for package in groups_map[group]["packages"]:
                group_packages_plain.add(package)
    return group_packages_dict, group_packages_plain


def get_packages(user, pkg_owners_map, groups_map):
    """
    Returns all packages owned by user (including those owned through a group)
    returns dict of:       "combined" - list of all packages, owned both directly or through group
                           "group" - dict (keys are group names) of packages owned only through group
                           "primary" - list of packages owned only directly
    """
    # packages_owners_map['rpms']['some_package_name'] contains list of 'some_package_name' maintainers
    if user == "orphan":  # blacklist orphan user which has lots of unnecessary packages and would choke up our servers
        return {
            "primary": [],
            "group": {},
            "combined": []
        }
    group_packages, group_packages_list = get_user_group_packages(user, groups_map)
    primary_packages = {package for package in pkg_owners_map['rpms'] if user in pkg_owners_map['rpms'][package]}
    if user in groups_map:
        primary_packages = set(groups_map[user]["packages"]) # User is group, return group packages as primary packages
    return {
        "primary": list(primary_packages),
        "group": group_packages,
        "combined": list(primary_packages | group_packages_list)  # Join sets of primary and group packages
    }


def get_package_versions(package):
    versions = {}
    releases = CACHE.get("fedora_releases")
    for release in releases["values"][:-1]: # Skip Rawhide
        versions["Fedora %d" % release] = {"stable": None, "testing": None}
    # Append Rawhide in standardized naming format
    versions["Fedora Rawhide"] = {"stable": None, "testing": None}
    for release in app.config["EPEL_RELEASES"]:
        versions[release_from_dist("el" + str(release))] = {"stable": None, "testing": None}

    bodhi_versions = get_json("https://src.fedoraproject.org/_dg/bodhi_updates/rpms/%s" % package)
    if not bodhi_versions:
        # Return dictionary we've prepared above containing empty fields for each release
        # versions = {"Fedora XX": {"stable": None, "testing": None}, "Fedora Rawhide": {"stable": None, "testing": None}...}
        return versions
    for release in bodhi_versions["updates"]:
        # Exclude ELN for now
        if "ELN" in release:
            continue
        formatted_release = release_from_dist(release.replace("-",""))
        try:
            versions[formatted_release]["stable"] = bodhi_versions["updates"][release]["stable"]
        except KeyError:
            pass
        try:
            versions[formatted_release]["testing"] = bodhi_versions["updates"][release]["testing"]
        except KeyError:
            pass

    # We need to query mdapi if we have None in Fedora stable values
    for release in versions:
        if "Fedora" in release and not versions[release]["stable"]:
            if release == "Fedora 31":
                # Skip F31 to avoid talking to mdapi too much
                # Won't be needed for later Fedoras as from F32, Bodhi was used even for Rawhide
                # TODO: Remove this hack after F31 EOL
                continue
            if "Rawhide" in release:
                # Before the mass rebuild, it'll be pretty common for packages to not have any Rawhide build
                # Let's copy out version from previous Fedora for cases when Rawhide does not have any build
                if versions["Fedora %d" % releases["values"][-2]]["stable"]:
                     versions[release]["stable"] = versions["Fedora %d" % releases["values"][-2]]["stable"]
                     # Such speed, much wow!
                     continue
                mdapi_response = get_json("https://mdapi.fedoraproject.org/%s/srcpkg/%s" % ("rawhide", package), False, False)
            else:
                mdapi_response = get_json("https://mdapi.fedoraproject.org/%s/srcpkg/%s" % ("f" + release[-2:], package), False, False)
            if not mdapi_response:
                continue
            versions[release]["stable"] = package + "-" + mdapi_response["version"] + "-" + mdapi_response["release"]

    return versions
